/*****************************************************************************************
 *
 *  This file is part of Portable Code::Blocks Launcher.
 *
 *  Copyright (C) 2009-2010 Biplab Kumar Modak
 *  Homepage: http://biplab.in
 *
 *  Modified by JLK, April 2010 and tentatively renumbered to 0.1.3
 *  - Portable CB fonts, tool path & icon added
 *  - Console now hidden in release builds
 *
 *  Portable Code::Blocks Launcher is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Portable Code::Blocks Launcher is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************************/
#include <cstdio>
#include <tchar.h>

#include "wchar_util.h"

/** @brief Modifies the system %APPDATA% environment variable to point to subdirectory
 * AppData of Code::Blocks.
 *
 * @param currDir WCHAR[] Launcher (and Code::Blocks) current directory
 * @return bool success (true) or failure (false)
 */
bool setAppData(WCHAR currentDir[])
{
    TCHAR currDir[MAX_PATH];
	WIN32_FIND_DATAW findData;
	HANDLE hFind;
    bool success = true;

    wcscpy(currDir, currentDir);
	wcscat(currDir, _T("\\AppData"));

	hFind = FindFirstFileW(_T("AppData"), &findData);
	if (hFind == INVALID_HANDLE_VALUE)
	{
	    fwprintf(stdout, _T("Creating AppData directory!\n"));
	    if (CreateDirectory(_T("AppData"), NULL) == 0)
            success = false;
	}
	FindClose(hFind);
	if (!success)
	    return false;

	hFind = FindFirstFileW(_T("AppData\\codeblocks"), &findData);
	if (hFind == INVALID_HANDLE_VALUE)
	{
	    fwprintf(stdout, _T("Creating AppData\\codeblocks directory!\n"));
	    if (CreateDirectory(_T("AppData\\codeblocks"), NULL) == 0)
           success = false;
	}
	FindClose(hFind);
	if (!success)
	    return false;

    return SetEnvironmentVariable(_T("APPDATA"), currDir);
}

/** @brief Add the tool subdirectory of Code::Blocks to the system path.
 *
 * @param currDir WCHAR[] Code::Blocks current directory
 * @return bool success (true) or failure (false)
 */
bool setToolPath(WCHAR currentDir[])
{
    const unsigned MAX_ENV = 4096; // 4k for %path% should do
	WIN32_FIND_DATAW findData;
	HANDLE hFind;
    bool success = true, ret = false;

	hFind = FindFirstFileW(_T("tool"), &findData);
	if (hFind == INVALID_HANDLE_VALUE)
	{
	    fwprintf(stdout, _T("Cannot find tool subdirectory!\n"));
        success = false;
	}
	FindClose(hFind);
	if (!success)
	    return false; // Give up if tool subdir does not exist

    // tool subdirectory exists. Add it and all its subdirectories to system path
    TCHAR toolDir[MAX_PATH], toFind[MAX_PATH];
    wcscpy(toolDir, currentDir);
	wcscat(toolDir, _T("\\tool"));
	wcscpy(toFind, toolDir);
	wcscat(toFind, _T("\\*.*"));

    TCHAR nuPath[MAX_ENV];
    wcscpy(nuPath, toolDir); // Add tools directory

    // Add all tool subdirectories
	hFind = FindFirstFile(toFind, &findData);
	if (hFind != INVALID_HANDLE_VALUE)
	{
        do // Add found file to path if a directory
        {
            if((findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
                && (wcscmp(findData.cFileName, _T(".")))
                && (wcscmp(findData.cFileName, _T(".."))))
            {
                wcscat(nuPath, _T(";"));
                wcscat(nuPath, toolDir);
                wcscat(nuPath, _T("\\"));
                wcscat(nuPath, findData.cFileName);
            }
            ret = FindNextFile(hFind, &findData);
        } while (ret);
	}
	FindClose(hFind);

    // Concatenate with existing system path
    TCHAR oldPath[MAX_ENV];
    if (GetEnvironmentVariable( _T("PATH"), oldPath, MAX_ENV))
    {
        wcscat(nuPath, _T(";"));
        wcscat(nuPath, oldPath);
    }
    fwprintf(stdout, _T("\nSetting path: %s\n"), nuPath);

    return SetEnvironmentVariable(_T("PATH"), nuPath);
}


/** @brief Add Code::Blocks fonts to system fonts if they exist in the given directory
 * - Assume fonts are available in the \share\CodeBlocks\fonts subdirectory of installation
 *
 * @param currDir WCHAR[] Launcher (and Code::Blocks) current directory
 * @return bool success (true) or failure (false)
 */
bool addFont(WCHAR currDir[])
{
	WIN32_FIND_DATA findData;
	HANDLE hFind;
	TCHAR fontsDir[MAX_PATH];
	TCHAR fontFile[MAX_PATH];
	int numFonts = 0, ret;

    wcscpy(fontsDir, currDir);
    wcscat(fontsDir, _T("\\share\\CodeBlocks\\fonts"));

    // Give up if no fonts directory exists
    hFind = FindFirstFile(fontsDir, &findData);
	if (hFind == INVALID_HANDLE_VALUE)
	    return false;
    FindClose(hFind);

    // Add CB fonts to system fonts
    wcscpy(fontFile, fontsDir);
    wcscat(fontFile, _T("\\*.ttf"));
	hFind = FindFirstFile(fontFile, &findData);
	if (hFind != INVALID_HANDLE_VALUE)
	{
        do
        {
            wcscpy(fontFile, fontsDir);
            wcscat(fontFile, _T("\\"));
            wcscat(fontFile, findData.cFileName);
            if (AddFontResource(fontFile))
            {
                fwprintf(stdout, _T("installed %s\n"), findData.cFileName);
                numFonts++;
            }
            else
                fwprintf(stderr, _T("installation of %s failed\n"), fontFile);

            ret = FindNextFile(hFind, &findData);
        } while (ret);
	}
	FindClose(hFind);

	// Broadcast the font change
	if(numFonts > 0)
	{
        SendMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
	    fwprintf(stdout, _T("installed %d fonts\n"), numFonts);
	    return true;
	}
	return false;
}

/** @brief Remove Code::Blocks fonts from system fonts if they were installed
 * - Assume fonts were added from the \share\CodeBlocks\fonts subdirectory of CB
 *
 * @param currDir WCHAR[] Launcher (and Code::Blocks) current directory
 */
void removeFont(WCHAR currDir[])
{
	WIN32_FIND_DATA findData;
	HANDLE hFind;
	TCHAR fontsDir[MAX_PATH];
	TCHAR fontFile[MAX_PATH];
	int ret;

    wcscpy(fontsDir, currDir);
    wcscat(fontsDir, _T("\\share\\CodeBlocks\\fonts"));

    // Give up if no fonts directory exists
    hFind = FindFirstFile(fontsDir, &findData);
	if (hFind == INVALID_HANDLE_VALUE)
	    return;
    FindClose(hFind);

    // Remove CB fonts from system fonts
    wcscpy(fontFile, fontsDir);
    wcscat(fontFile, _T("\\*.ttf"));
	hFind = FindFirstFile(fontFile, &findData);
	if (hFind != INVALID_HANDLE_VALUE)
	{
        do
        {
            wcscpy(fontFile, fontsDir);
            wcscat(fontFile, _T("\\"));
            wcscat(fontFile, findData.cFileName);
            RemoveFontResource(fontFile);
            ret = FindNextFile(hFind, &findData);
        } while (ret);
	}
	FindClose(hFind);

	// Broadcast the font change
    SendMessage(HWND_BROADCAST, WM_FONTCHANGE, 0, 0);
	fwprintf(stdout, _T("removed Code::Blocks extra fonts\n"));
}


int main()
{
   	bool fontInstalled = false;
    TCHAR currDir[MAX_PATH];

    // Console output will show only in debug build target
    wprintf(_T("***********************************************************\n"));
    wprintf(_T("*          Portable Code::Blocks Launcher v0.1.3          *\n"));
    wprintf(_T("*         Copyright (C) 2010 - Biplab Kumar Modak         *\n"));
    wprintf(_T("*               Homepage: http://biplab.in                *\n"));
    wprintf(_T("*                   License: GPL v3                       *\n"));
    wprintf(_T("***********************************************************\n"));


    if (! GetCurrentDirectory(MAX_PATH, currDir))
		fwprintf(stderr, _T("Couldn't retrieve current directory!\n"));

    // Add application fonts to system
    fontInstalled = addFont(currDir);

    // Set APPDATA environment variable
    if (!setAppData(currDir))
	{
        fwprintf(stderr, _T("Unable to set APPDATA environment variable\n"));
		return 1;
	}

    // Add tools subdirectories to %path%
    if (!setToolPath(currDir))
        fwprintf(stderr, _T("Unable to set PATH environment variable\n"));

	// Launch Code::Blocks
	STARTUPINFO si;
	PROCESS_INFORMATION pi;

	ZeroMemory(&si, sizeof(si));
	ZeroMemory(&pi, sizeof(pi));
	si.cb = sizeof(si);
    wprintf(_T("\nLaunching Code::Blocks... "));

	if (! CreateProcess(_T("codeblocks.exe"),
                        toWCHAR("codeblocks.exe --debug-log"),
                        NULL, NULL,
                        FALSE, DETACHED_PROCESS, NULL, NULL,
                        &si, &pi ))
	{
        fwprintf(stderr, _T("\nUnable to launch codeblocks.exe\n"));
		return 1;
	}

    // Wait until CB terminates
	wprintf(_T("Success\n"));
	WaitForSingleObject(pi.hProcess, INFINITE);

    // Remove portable CB fonts from system if installed
    if (fontInstalled)
        removeFont(currDir);

    return 0;
}
